#---------------------------------------------------------------------------------------
# Singapore, 1st July 2017

# ACADEMIC SOFTWARE LICENSE for PLSDBC
# ************************************

# Copyright:
# Wing-Cheong Wong
# Bioinformatics Institute (BII) A*STAR Singapore

# This software is subjected to license restrictions. The software is
# provided as is. In its present form and without written consent of 
# the copyright holder, the software is provided in accordance with GPL
# (www.gnu.org/licenses/gpl.html). Among other issues, this excludes any
# warranty and indemnity when using this software. For a commercial license,
# please approach the authors (via email to wongwc@bii.a-star.edu.sg).

#---------------------------------------------------------------------------------------

#!/usr/bin/perl
package biascorrAlgo;

use strict;	# Forces variables to be declared
use warnings;
use POSIX;
use biascorrUtil;

# Normalize (through scaling in log-log scale) count data within the experiment to Zipf's law (theoretical distribution for gene expression)
sub scaleDatatoZipfLaw {
	my ($dataRef, $regressRef, $maxcount, $slopeuser, $base) = @_;
	my @regressionlist=@$regressRef;
	my @data=@$dataRef;

	my $j; my $k;
	my $var1; my $var2;
	my @tmp1; my @tmp2; my @tmp3; my @tmp4;
	my $count; my $exptidx;
	my $slope; my $intercept;
	my $logbase=log($base);
	my $x1; my $x2;

	# Create a matrix of the same size as data array to reflect the status of each datapoint
	my @statusMat=(); my $status;
	for ($j=0; $j<scalar(@data); $j++) {
		@tmp1=split(/\t/,$data[$j]);
		@tmp2=(); push(@tmp2,$tmp1[0]);	# Initialize with transcript name
		for ($k=1; $k<scalar(@tmp1); $k++) { push(@tmp2,0); }	# Initialize each datapoint across experiments
		$var1=join("\t", @tmp2);
		push(@statusMat,$var1);
	}

	for ($j=0; $j<scalar(@regressionlist); $j++) {
		@tmp1 = split(/\;/, $regressionlist[$j]);	# e.g. 0;0;5.603,6.603;-0.97;6.51;0.96;11 i.e. numexpt;intervalID;range;slope;intercept;Rsq;nsize
		$exptidx = $tmp1[0]+1;						# Get experiment column wrt data array
		@tmp2 = split(/\,/,$tmp1[2]); $var1=$tmp2[0]; $var2=$tmp2[1];	# Get the interval range

		# if it is the first segment, increase the upper limit margin to capture the largest point in cases of rounding error
#		if ($tmp1[1]==0) { $var2=$var2+0.5; }
		if ($tmp1[1]==0) { $var2=$var2*2; }

		# if it is the last segment, decrease the lower limit margin to capture the smallest point in cases of rounding error
		if ($tmp1[1]==scalar(@regressionlist)-1) { $var1=0; }
		print "Segment[".$j."] range:\t".$var2.",".$var1."\n";

		# Scaling parameter for this experiment
		$slope = $slopeuser;	# Theoretical slope of -1

		$intercept = log($maxcount)/$logbase;						# The maximum count for the experiment
#		$intercept = log($maxcountPerExpt[$tmp1[0]])/$logbase;		# The maximum count for the experiment
#		$intercept = log($maxcountPerExpt[$refexptidx])/$logbase;	# The maximum count among all experiments

#		print "Transformed output[size/slope/intercept]: ".$regressionlist[$j]."\t".scalar(@keys)."\t".$slope.";".$intercept.";".$tmp1[4].";".$tmp1[3]."\n";

#		%hashtable1=()
		for ($k=0; $k<scalar(@data); $k++) {
			# 'Per transcript count' as the key of hashtable, 'Name of transcripts with this count' as the value of hashtable
			@tmp3=split(/\t/,$data[$k]);

			# Convert the transcript count to logarithmic scale
			$count=log($tmp3[$exptidx])/$logbase;

			# Check if the data point has been corrected previously to prevent it being corrected again if it falls into the range of the next interval
			@tmp4=split(/\t/,$statusMat[$k]);
			$status=$tmp4[$exptidx];

			if (($count<=$var2 && $count>$var1)&&$status==0) {	# Check if datapoint is within the interval and that it has not been previously corrected
#				print Before."[".$k."][".$exptidx."]\t".$data[$k]."\t".$statusMat[$k]."\n";
				if ($tmp3[$exptidx]>1) {	# Exclude adjustment for pseudocounts

#					$x1=(($base**$tmp1[4]/$base**$intercept)**(1/$slope));
#					$x2=($tmp3[$exptidx]**($tmp1[3]/$slope));
#					print $tmp3[$exptidx].";".$intercept.";".$slope.";".$tmp1[4].";".$tmp1[3].";".$x1.";".$x2.";";

					# Perform the slope conversion. Note that $tmp1[4] is the original intercept while $tmp1[3] is the original slope of this log-log segment
					$tmp3[$exptidx]=floor((($base**$tmp1[4]/$base**$intercept)**(1/$slope))*($tmp3[$exptidx]**($tmp1[3]/$slope)));
					if ($tmp3[$exptidx]<2) { $tmp3[$exptidx]=2; }	# Constraint the final value to be at least 2 (1 above pseudo count)
					$data[$k]=join("\t",@tmp3);

					# Set the status of the datapoint and update the status matrix
					$tmp4[$exptidx]=1; $statusMat[$k]=join("\t", @tmp4);

#					print $tmp3[$exptidx]."\n";
#					print After."[".$k."][".$exptidx."]\t".$data[$k]."\t".$statusMat[$k]."\n";
				}
			}
		}
	}
	return \@data;
}

sub computeOptimumShiftfactor {
	my ($regresslistref, $dataRef, $logbase) = @_;

	my @regressionlist = @$regresslistref;
	my @data=@$dataRef;	# Assumes data comes with header

	my $i;
	my $j;
	my $k;
	my $var1;
	my $var2;
	my $var3;
	my @tmp1;
	my @tmp2;
	my @tmp3;
	my $numexpt;
	my @intervallist1=(); my @regressionlist1=();
	my @intervallist2=(); my @regressionlist2=();
	my @intervallist3=(); my @regressionlist3=();
	my @intervallist4=(); my @regressionlist4=();
	my @optregressionlist=();

	# Compute the number of experiments
	@tmp1=split(/\t/, $data[0]); $numexpt=scalar(@tmp1)-1;

	# Create interval lists based on the various excluded segments per experiment
	for ($k=0; $k<$numexpt; $k++) {
		for ($i=0,@tmp3=(); $i<scalar(@regressionlist); $i++) {
			@tmp1 = split(/\;/, $regressionlist[$i]);	# numexpt;intervalID;range;slope;intercept;Rsq;nsize
			if ($tmp1[0]==$k) { @tmp2 = split(/\,/, $tmp1[2]); push(@tmp3,$tmp2[0]); push(@tmp3,$tmp2[1]); }
		}
		@tmp1 = sort {$a<=>$b} @tmp3;

		# Include all segments
		push(@intervallist1, $k.";0;".$tmp1[0].",".$tmp1[scalar(@tmp1)-1]);

		# Exclude first segment (largest values); Need at least 2 segments
		if (scalar(@tmp1)>2) { push(@intervallist2, $k.";0;".$tmp1[0].",".$tmp1[scalar(@tmp1)-3]); }
		else { push(@intervallist2, $k.";0;".$tmp1[0].",".$tmp1[scalar(@tmp1)-1]); }

		# Exclude first and last segments (largest and smallest values); Need at least 3 segments
		if (scalar(@tmp1)>6) { push(@intervallist3, $k.";0;".$tmp1[2].",".$tmp1[scalar(@tmp1)-3]); }
		else { push(@intervallist3, $k.";0;".$tmp1[0].",".$tmp1[scalar(@tmp1)-1]); }

		# Exclude first,second and last segments (largest and smallest values); Need at least 4 segments
		if (scalar(@tmp1)>8) { push(@intervallist4, $k.";0;".$tmp1[4].",".$tmp1[scalar(@tmp1)-3]); }
		else { push(@intervallist4, $k.";0;".$tmp1[0].",".$tmp1[scalar(@tmp1)-1]); }

	}
	# Perform regression for the following intervals
	($var1,$var2)=computeRegressionParameters(\@data, $numexpt, \@intervallist1, $logbase); @regressionlist1=@$var1; @intervallist1=@$var2;	# type 1: all segments in sample
	($var1,$var2)=computeRegressionParameters(\@data, $numexpt, \@intervallist2, $logbase); @regressionlist2=@$var1; @intervallist2=@$var2;	# type 2: exclude 1st segment in sample
	($var1,$var2)=computeRegressionParameters(\@data, $numexpt, \@intervallist3, $logbase); @regressionlist3=@$var1; @intervallist3=@$var2;	# type 3: exclude 1st and last segments in sample
	($var1,$var2)=computeRegressionParameters(\@data, $numexpt, \@intervallist4, $logbase); @regressionlist4=@$var1; @intervallist4=@$var2;	# type 3: exclude 1st,2nd and last segments in sample

	my @abserror=(); my $minidx=-1; my $minerror;
	for ($k=0; $k<$numexpt; $k++) {
		@tmp1=();
		push(@tmp1, $regressionlist1[$k]);
		push(@tmp1, $regressionlist2[$k]);
		push(@tmp1, $regressionlist3[$k]);
		push(@tmp1, $regressionlist4[$k]);

		# Compute the total slope (ideal slope=-1) and Rsq (ideal Rsq=1)error for each interval type
		for ($i=0,@abserror=(); $i<scalar(@tmp1); $i++) {
			@tmp2 = split(/\;/, $tmp1[$i]);	# numexpt;intervalID;range;slope;intercept;Rsq;nsize
			push(@abserror,abs(-1.0-$tmp2[3]) + abs(1.0-$tmp2[5]));
		}
		for ($i=0,$minerror=$abserror[0],$minidx=0; $i<scalar(@abserror); $i++) {
			if ($minerror > $abserror[$i]) { $minerror=$abserror[$i]; $minidx=$i; }
		}
		push(@optregressionlist,$tmp1[$minidx]);
#		print $regressionlist1[$k]."\n".$regressionlist2[$k]."\n".$regressionlist3[$k]."\n";
#		print $k."\t".$minidx."\t".$minerror."\t".$tmp1[$minidx]."\n";
	}
	return \@optregressionlist;
}

sub computeRegressionParameters {
	my ($dataRef, $numexpt, $intervalsperExptRef, $logbase) = @_;
	my @data=@$dataRef;
	my @intervalsperExpt=@$intervalsperExptRef;

	my $i;
	my $j;
	my $var1;
	my $var2;
	my $var3;
	my @tmp1;
	my @tmp2;
	my @tmp3;
	my @tmp4;
	my @tmp5;
	my $slope; my $intercept; my $Rsq; my $stdslope;
	my @Xlist=(); my @Ylist=();
	my $avgX; my $stddevX; my $avgY; my $stddevY;
	my @regressionlist=();	# return results

	# Get the log-log datapoints for each interval
	($var1, $var2)=biascorrUtil::getIntervaldatapts(\@data, $numexpt, \@intervalsperExpt, $logbase);
	@Xlist=@$var1; @Ylist=@$var2;

	# Perform regression analysis for each log-log interval
	if (scalar(@intervalsperExpt)==scalar(@Xlist) && scalar(@intervalsperExpt)==scalar(@Ylist) && scalar(@Xlist)>0) {

		# Join the last two segments if the last segment has less than N points
#		print "Before\n"; for ($j=0; $j<scalar(@intervalsperExpt); $j++) { print $intervalsperExpt[$j]."\n"; } print "\n";
		@tmp1 = split(/\;/, $Xlist[scalar(@Xlist)-1]);
		if (scalar(@tmp1)<$configs::REGRESSMINPTS) {
			$Xlist[scalar(@Xlist)-2]=$Xlist[scalar(@Xlist)-2].";".$Xlist[scalar(@Xlist)-1];
			$Ylist[scalar(@Ylist)-2]=$Ylist[scalar(@Ylist)-2].";".$Ylist[scalar(@Ylist)-1];

			# Compute the new range of the second last segment
			@tmp2 = split(/\;/, $intervalsperExpt[scalar(@intervalsperExpt)-2]); @tmp2 = split(/\,/, $tmp2[2]); # 0:exptid;1:segid;2:lower,upper
			@tmp3 = split(/\;/, $intervalsperExpt[scalar(@intervalsperExpt)-1]); @tmp3 = split(/\,/, $tmp3[2]); # 0:exptid;1:segid;2:lower,upper

			@tmp4 = split(/\;/, $intervalsperExpt[scalar(@intervalsperExpt)-2]);
			$intervalsperExpt[scalar(@intervalsperExpt)-2]=$tmp4[0].";".$tmp4[1].";".$tmp3[0].",".$tmp2[1];

			# Delete last segment intervalsperExpt and datapoints
			$var1 = pop @intervalsperExpt;
			$var2 = pop @Xlist;
			$var3 = pop @Ylist;
		}
#		print "After\n"; for ($j=0; $j<scalar(@intervalsperExpt); $j++) { print $intervalsperExpt[$j]."\n"; } print "\n";

		for ($j=0,@regressionlist=(); $j<scalar(@Xlist); $j++) {
			@tmp1 = split(/\;/, $Xlist[$j]);
			@tmp2 = split(/\;/, $Ylist[$j]);

			($slope, $intercept, $Rsq, $stdslope) = mathsfun::linearRegression(\@tmp1, \@tmp2, 'INTERCEPT');
			if (!($slope ne "" && $intercept ne "" && $Rsq ne "")) { $slope=0; $intercept=0; $Rsq=-1; $stdslope=0; }

#			if ($slope ne "" && $intercept ne "" && $Rsq ne "") {
#				$slope=sprintf("%.5f",$slope); $intercept=sprintf("%.5f",$intercept); $Rsq=sprintf("%.5f",$Rsq);
#			}
#			else { $slope=0; $intercept=0; $Rsq=-1; }
#			print $intervalsperExpt[$j].";".$slope.";".$intercept.";".$Rsq.";".scalar(@tmp1)."\n";
			push(@regressionlist, $intervalsperExpt[$j].";".$slope.";".$intercept.";".$Rsq.";".scalar(@tmp1));
		}
	}
	return (\@regressionlist, \@intervalsperExpt);
}

1;