#---------------------------------------------------------------------------------------
# Singapore, 1st July 2017

# ACADEMIC SOFTWARE LICENSE for PLSDBC
# ************************************

# Copyright:
# Wing-Cheong Wong
# Bioinformatics Institute (BII) A*STAR Singapore

# This software is subjected to license restrictions. The software is
# provided as is. In its present form and without written consent of 
# the copyright holder, the software is provided in accordance with GPL
# (www.gnu.org/licenses/gpl.html). Among other issues, this excludes any
# warranty and indemnity when using this software. For a commercial license,
# please approach the authors (via email to wongwc@bii.a-star.edu.sg).

#---------------------------------------------------------------------------------------

#!/usr/bin/perl
package mathsfun;

use strict;	# Forces variables to be declared
use warnings;
use POSIX;

# Maths constant definition
our $PI = 3.141592653589793;	# 15 decimal places

sub BinomialCoefficient {
	my $i;
	my $results;
	my($n, $k) = @_;

	if ($k > 0) {
		for ($i=1; $i<=$k; $i++) {
			if ($i == 1) {	$results = ($n-($k-$i))/$i; }
			else { $results = $results * ($n-($k-$i))/$i; }
		}
	}
	else { $results = 1; }
	return $results;
}

sub BinomialPDF {
	my($n, $k, $p) = @_;
	my $results;
	if ($n>0) { $results= BinomialCoefficient($n, $k) * ($p ** $k) * ((1.0 - $p)**($n-$k)); }
	else { $results = 0.0; }
	return $results;
}

sub BinomialCDF {
	my $x;
	my $results;
	my($n, $k, $p) = @_;

	for ($x=0,$results=0; $x<=$k; $x++) {
		if ($x==0) { $results = BinomialPDF($n, $x, $p); }
		else { $results = $results + BinomialPDF($n, $x, $p); }
	}
	return $results;
}

sub erf { # Maclaurin series
	my($z) = @_;

	my $a = 1/3;
	my $b = 1/10;
	my $c = 1/42;
	my $d = 1/216;
	my $e = 1/1320;
	my $f = 1/9360;
	my $results = (2/sqrt($PI)) * ($z - $a*($z**3) + $b*($z**5) - $c*($z**7) + $d*($z**9) - $e*($z**11) + $f*($z**13));
#	my $results = (2/sqrt($PI)) * ($z - $a*($z**3) + $b*($z**5) - $c*($z**7) + $d*($z**9));
	return $results;
}

sub normalCDF {
	my($mean, $var, $x) = @_;

	my $z = ($x-$mean)/sqrt(2*$var);
	my $results = 0.5*(1 + erf($z));

	if ($results > 1) { $results = 1.0; }
	if ($results < 0) { $results = 0.0; }

	return $results;
}

sub hypergeometricCDF {
	my ($R1,$f11,$N,$C1) = @_;

	my $i;
	my $var1;
	my $sumprob; my $totalprob;
	my $result=-1;

	if ($R1 =~ /^-?\d+\.?\d*$/ && $f11 =~ /^-?\d+\.?\d*$/ && $N =~ /^-?\d+\.?\d*$/ && $C1 =~ /^-?\d+\.?\d*$/) { # check if it is a real number
		for ($i=0,$totalprob=0,$sumprob=0; $i<=$N; $i++) {
			$var1 = hypergeometricPDF($R1,$i,$N,$C1);
			if ($var1 >= 0) {
				if ($i<=$f11) { $sumprob = $sumprob + $var1; }	# sum up to k where P(X=k)
				$totalprob = $totalprob + $var1;		# sum for all probabilities since rounding errors might result in total probability exceed 1
#				print "P(X=".$i.")=".$var1."\n";
			}
		}
		if ($totalprob>=$sumprob && $totalprob>0) { $result = $sumprob/$totalprob; }
		else { $result = -1; }
#		print $totalprob."\t".$sumprob."\t".$result."\n";
	}
	else { $result = -1; }	# returns negative value to denote error

	return $result;
}

sub hypergeometricPDF {
#	 2 by 2 contingency table i.e. P(X=f11) = (R1 f11)(R2 f22)(N C1)
#	 f11 f12|R1
#	 f21 f22|R2
#	 ----------
#	  C1  C2| N
	my ($R1,$f11,$N,$C1) = @_;
	my $var1; my $var2; my $var3;
	my $R2; my $C2;
	my $f12; my $f21; my $f22;
	my $result=-1;
	my $ans1; my $ans2;

	if ($R1 =~ /^-?\d+\.?\d*$/ && $f11 =~ /^-?\d+\.?\d*$/ && $N =~ /^-?\d+\.?\d*$/ && $C1 =~ /^-?\d+\.?\d*$/) { # check if it is a real number
		$R2 = $N-$R1; $C2 = $N-$C1;
		$f12 = $R1-$f11; $f21 = $C1-$f11; $f22 = $R2-$f21;
#		print $f11."\t".$f12."\t|".$R1."\n".$f21."\t".$f22."\t|".$R2."\n".$C1."\t".$C2."\t|".$N."\n";

		# Compute the probability in log scale
		if ($f11>=0 && $f12>=0 && $f21>=0 && $f22>=0) {

			$var1 = BinomialCoefficient($R1,$f11);
			$var2 = BinomialCoefficient($R2,$f21);
			$var3 = BinomialCoefficient($N,$C1);
			if ($var1 =~ /^-?\d+\.?\d*$/ && $var2 =~ /^-?\d+\.?\d*$/ && $var3 =~ /^-?\d+\.?\d*$/) {
				$ans1 = ($var1/$var3)*$var2;
			} else { $ans1 = 'nan'; }
#			print "[1]:\t".$var1.",".$var2.",".$var3.",".$ans1."\n";

			$var1 = logNchooseK($R1,$f11);
			$var2 = logNchooseK($R2,$f21);
			$var3 = logNchooseK($N,$C1);
			$ans2 = $var1 + $var2 - $var3;
			$ans2 = exp($ans2);
#			print "[2]:\t".$var1.",".$var2.",".$var3.",".$ans2."\n";

			if ($ans1 ne 'nan' && $ans1>=0 && $ans1<=1) { $result = $ans1; }
			elsif ($ans2 ne 'nan' && $ans2>=0 && $ans2<=1) { $result = $ans2; }
			else { $result = -1; }	# returns negative value to denote error
		}
		else { $result = -1; }	# returns negative value to denote error
	}
	return $result;
}

# Using Stiring's approximation for log factorial
sub logNchooseK {
	my ($N, $K) = @_;

	my $i;
	my $result;
	my $var1; my $var2; my $var3; my $var4;

	if ($N =~ /^-?\d+\.?\d*$/ && $K =~ /^-?\d+\.?\d*$/ && $N>=0 && $K>=0) { # check if it is a real number
		if ($N>$K && $K>0) {
			$var1 = ($N+0.5)*log($N);
			$var2 = ($K+0.5)*log($K);
			$var3 = ($N-$K+0.5)*log($N-$K);
			$result = $var1 - $var2 - $var3 - 0.5*log(2*$PI);
			$var4 = exp($result);
		}
		else { $result=0; }	# (N N)=1;(N 0)=1;(0 0)=1. And in terms of log scale, return 0
	}
	else { $result = -1; }

	return $result;
}

sub factorial {
	my ($number) = @_;

	my $i;
	my $logsum=0;
	my $result;

	if ($number =~ /^-?\d+\.?\d*$/) { # check if it is a real number
		# Compute the factorial x! as sum of log terms i.e log(x!)
		for ($i=1,$logsum=0; $i<=$number; $i++) { $logsum = $logsum + log($i); }
	}
	# Convert the log scale back to original scale
	if ($logsum > 0) { $result = exp($logsum); }
	else { $result = 1; }

	return floor($result);
}

sub logsum {
	my($arrayRef1, $arrayRef2) = @_;
	my @array1 = @$arrayRef1;
	my @array2 = @$arrayRef2;

	my $i;
	my $var;
	my @results;
	if (scalar(@array1)==scalar(@array2) && scalar(@array1)>0) {
		for ($i=0,@results=(); $i<scalar(@array1); $i++) {
			if ($array1[$i] =~ /^-?\d+\.?\d*$/ && $array2[$i] =~ /^-?\d+\.?\d*$/) { # check if it is a real number
				$var = log($array1[$i]) + log($array2[$i]);
				$var = sprintf('%.10f', $var);
			} else { $var = 'x'; }
			push(@results, $var);
		}
	}
	else { @results = (); }
	return \@results;
}

# See Jerrold Zar, Biostatistical analysis, Chapter 17.3, pp326-335
sub linearRegression {
	my($arrayRefX, $arrayRefY, $type) = @_;
	my @arrayA = @$arrayRefX;
	my @arrayB = @$arrayRefY;

	my $i;
	my $var1; my $var2;
	my @results;

	# Ensure that the arrays contain valid numbers
	my @arrayX=(); my @arrayY=();
	if (scalar(@arrayA)==scalar(@arrayB) && scalar(@arrayA)>0) {
		for ($i=0; $i<scalar(@arrayA); $i++) {
#			print "Datapoints:".$arrayA[$i]."\t".$arrayB[$i]."\n";
			if ($arrayA[$i] =~ /^-?\d+\.?\d*$/ && $arrayB[$i] =~ /^-?\d+\.?\d*$/) { # check if it is a real number
				push(@arrayX, $arrayA[$i]); push(@arrayY, $arrayB[$i]);
			}
		}
	}
#	print scalar(@arrayX)."\t".scalar(@arrayY)."\n";

	my $sumXY=0; my $sumX=0; my $sumY=0; my $sumXsq=0; my $sumYsq=0; my $n;
	my $totalSS; my $regressionSS; my $residualSS;
	my $totalDF; my $regressionDF; my $residualDF;
	my $regressionMS; my $residualMS; my $F;
	my $slope; my $intercept; my $Rsq;
	my $stdslope;
	if (scalar(@arrayX)==scalar(@arrayY) && scalar(@arrayX)>1) {	# At least 2 datapoints
		$n = scalar(@arrayX);
		$totalDF=$n; $regressionDF=1; $residualDF=$totalDF-$regressionDF;
#		for ($i=0; $i<scalar(@arrayX); $i++) { print $i."\t".$arrayX[$i]."\t".$arrayY[$i]."\n"; }

		for ($i=0; $i<scalar(@arrayX); $i++) {
			$sumXY = $sumXY + ($arrayX[$i]*$arrayY[$i]);
			$sumX = $sumX + $arrayX[$i];
			$sumY = $sumY + $arrayY[$i];
			$sumXsq = $sumXsq + ($arrayX[$i]*$arrayX[$i]);
			$sumYsq = $sumYsq + ($arrayY[$i]*$arrayY[$i]);
		}
		if ($type eq 'INTERCEPT') {
			$slope = ($sumXY - (($sumX*$sumY)/$n)) / ($sumXsq-(($sumX*$sumX)/$n));
			$intercept = $sumY/$n - $slope*($sumX/$n);
		}
		else { $slope = $sumXY/$sumXsq;	$intercept = 0;	}

		$totalSS = $sumYsq - ($sumY*$sumY)/$n;	# (17.10)
		if (($sumXsq-(($sumX*$sumX)/$n))>0 && $totalSS>0) {
			$regressionSS = (($sumXY-(($sumX*$sumY)/$n))**2) / ($sumXsq-(($sumX*$sumX)/$n));	# (17.11)
			$residualSS=$totalSS-$regressionSS;	# (17.13)
			$regressionMS=$regressionSS/$regressionDF; $residualMS=$residualSS/$residualDF;	$F=$regressionMS/$residualMS;	# (17.14)
			$Rsq = $regressionSS/$totalSS;	# (17.15)
			$stdslope=sqrt($residualMS/$sumXsq);	# (17.20)
		}
		else { $regressionSS=""; $residualSS=""; $Rsq=""; $regressionMS=""; $residualMS=""; $F=""; }

#		print "Linear regression: ".$totalSS.",".$regressionSS.",".$n.",".$Rsq."\n";
#		print $slope."\t".$intercept."\t".$Rsq."\n";
	}
	else { 	$slope=""; $intercept=""; $Rsq=""; $stdslope=""; }

	return ($slope, $intercept, $Rsq, $stdslope);
}

# See Jerrold Zar, Biostatistical analysis, Chapter 18.4(pp370-372) and Chapter 17.3(334-335)
sub testmultipleslopes {
	my($arrayXRef, $arrayYRef, $statsTable) = @_;
	my @Xlist = @$arrayXRef;
	my @Ylist = @$arrayYRef;

	my $i;
	my $j;
	my $k;
	my $var1;
	my $var2;
	my $line;
	my @tmp1;
	my $messageflag=1;
	my $status="PASS";

	my @dataX; my @dataY; my $numexpt;

	my $sumXsq; my $sumX; my $sumYsq; my $sumY; my $sumXY;	# variables for per regression line
	my $totalSS; my $regressionSS; my $residualSS; my $residualDF;
	my $slope; my $intercept; my $Rsquare;

	my $sumXsqtotal; my $sumXtotal; my $sumYsqtotal; my $sumYtotal; my $sumXYtotal;	# variables for total regression line
	my $totalSStotal; my $regressionSStotal; my $residualSStotal; my $residualDFtotal;
	my $slopetotal; my $intercepttotal; my $Rsquaretotal;

	my $sumXsqcommon; my $sumXYcommon; my $sumYsqcommon; my $residualSScommon; my $residualDFcommon; my $slopecommon; my $totalsize;	# variables for common regression

	my $residualSSpooled; my $residualDFpooled;	# variables for pooled regression

	my $Fstatistics=""; my $significantflag="";
	my $numeratorDF=""; my $denominatorDF=""; my $Fcritical="";
	my $mindist; my $minidxCol; my $minidxRow;

	if (scalar(@Xlist)==scalar(@Ylist) && scalar(@Xlist)>1) {
		if ($messageflag==1) {
			print "Multiple testing report:\n";
			print "Type\tSumXsq\tSumXY\tSumYsq\tResidualSS\tResidualDF\tRegression parameters[Slope/Intercept/Rsq]\n";
		}

		$residualSSpooled=0; $residualDFpooled=0;
		$residualSScommon=0; $residualDFcommon=0;
		$sumXsqcommon=0; $sumXYcommon=0; $sumYsqcommon=0;
		$numexpt = scalar(@Xlist);

		$sumXsqtotal=0; $sumXtotal=0; $sumYsqtotal=0; $sumYtotal=0; $sumXYtotal=0; $totalsize=0;

		for ($i=0; $i<$numexpt; $i++) {
			@dataX = split(/\;/, $Xlist[$i]);
			@dataY = split(/\;/, $Ylist[$i]);

			$sumXsq=0; $sumX=0; $sumYsq=0; $sumY=0; $sumXY=0;
			$residualSS=0; $residualDF=0;
			for ($j=0; $j<scalar(@dataX); $j++) {
				$sumXsq = $sumXsq + ($dataX[$j]*$dataX[$j]);	# For per line regression
				$sumX = $sumX + $dataX[$j];
				$sumYsq = $sumYsq + ($dataY[$j]*$dataY[$j]);
				$sumY = $sumY + $dataY[$j];
				$sumXY = $sumXY + ($dataX[$j]*$dataY[$j]);

				$sumXsqtotal = $sumXsqtotal + ($dataX[$j]*$dataX[$j]);	# For total regression
				$sumXtotal = $sumXtotal + $dataX[$j];
				$sumYsqtotal = $sumYsqtotal + ($dataY[$j]*$dataY[$j]);
				$sumYtotal = $sumYtotal + $dataY[$j];
				$sumXYtotal = $sumXYtotal + ($dataX[$j]*$dataY[$j]);
				$totalsize = $totalsize + 1;
			}

			# Compute for per line regression
			$totalSS = $sumYsq-(($sumY*$sumY)/scalar(@dataY));

			$var1 = $sumXY-(($sumX*$sumY)/scalar(@dataX));
			$var2 = $sumXsq-(($sumX*$sumX)/scalar(@dataX));
			if ($var2==0) { $status="FAIL"; last; }	# Checkpoint to avoid division by zero
			$regressionSS = ($var1**2)/$var2;

			$residualSS = $totalSS-$regressionSS;
			$residualDF = scalar(@dataX)-2;

			$slope = ($sumXY-($sumX*$sumY)/scalar(@dataX))/($sumXsq-(($sumX*$sumX)/scalar(@dataX)));
			$intercept = $sumY/scalar(@dataX) - $slope*($sumX/scalar(@dataX));
			if ($totalSS==0) { $status="FAIL"; last; } 	# Checkpoint to avoid division by zero
			$Rsquare=$regressionSS/$totalSS;

			# Compute common sum of Xsquare, Ysquare and XY for common regression
			$sumXYcommon = $sumXYcommon + $var1;
			$sumXsqcommon = $sumXsqcommon + $var2;
			$sumYsqcommon = $sumYsqcommon + $totalSS;
			$residualDFcommon = $residualDFcommon + scalar(@dataX);

			# Compute pooled residual SS and DF for pooled regression
			$residualSSpooled = $residualSSpooled + $residualSS;
			$residualDFpooled = $residualDFpooled + $residualDF;

			if ($messageflag==1) {
#				$sumXsq = sprintf("%.2f",$sumXsq); $sumXY = sprintf("%.2f",$sumXY); $sumYsq = sprintf("%.2f",$sumYsq);
				$var2=sprintf("%.2f",$var2); $var1=sprintf("%.2f",$var1); $totalSS = sprintf("%.2f",$totalSS);
				$residualSS = sprintf("%.2f",$residualSS); $residualDF = sprintf("%d",$residualDF);
				$slope = sprintf("%.2f",$slope); $intercept = sprintf("%.2f",$intercept); $Rsquare = sprintf("%.2f",$Rsquare);
#				print "Regression".$i."\t".$sumXsq."\t".$sumXY."\t".$sumYsq."\t".$residualSS."\t".$residualDF."\t".$slope.";".$intercept.";".$Rsquare."\n";
				print "Regression".$i."\t".$var2."\t".$var1."\t".$totalSS."\t".$residualSS."\t".$residualDF."\t".$slope.";".$intercept.";".$Rsquare."\n";
			}
		}

		# Required values are valid i.e., no division by zero has occurred
		if ($status eq "PASS") {

			# For common regression
			$residualSScommon = $sumYsqcommon - (($sumXYcommon*$sumXYcommon)/$sumXsqcommon);
			$residualDFcommon = $residualDFcommon-$numexpt-1;
			$slopecommon = $sumXYcommon/$sumXsqcommon;

			# For total regression
			$totalSStotal = $sumYsqtotal-(($sumYtotal*$sumYtotal)/$totalsize);

			$var1 = $sumXYtotal-(($sumXtotal*$sumYtotal)/$totalsize);
			$var2 = $sumXsqtotal-(($sumXtotal*$sumXtotal)/$totalsize);
			$regressionSStotal = ($var1**2)/$var2;

			$residualSStotal = $totalSStotal-$regressionSStotal;
			$residualDFtotal = $totalsize-2;

			$slopetotal = ($sumXYtotal-($sumXtotal*$sumYtotal)/$totalsize)/($sumXsqtotal-(($sumXtotal*$sumXtotal)/$totalsize));
			$intercepttotal = $sumYtotal/$totalsize - $slopetotal*($sumXtotal/$totalsize);
			$Rsquaretotal=$regressionSStotal/$totalSStotal;

			# Compute the F statistics
			if ($residualDFpooled > 0) {
				$Fstatistics = (($residualSScommon-$residualSSpooled)/($numexpt-1)) / ($residualSSpooled/$residualDFpooled);	# Eqn 18.29 (see Jar)
#				$Fstatistics = (($residualSStotal-$residualSSpooled)/(2*($numexpt-1))) / ($residualSSpooled/$residualDFpooled);	# Eqn 18.49 (see Jar)
			}
			else { $Fstatistics="x"; }

			if ($messageflag==1) {
				$residualSSpooled = sprintf("%.2f",$residualSSpooled); $residualDFpooled = sprintf("%d",$residualDFpooled);
				$sumXsqcommon = sprintf("%.2f",$sumXsqcommon); $sumXYcommon = sprintf("%.2f",$sumXYcommon); $sumYsqcommon = sprintf("%.2f",$sumYsqcommon);
				$residualSScommon = sprintf("%.2f",$residualSScommon); $residualDFcommon = sprintf("%d",$residualDFcommon); $slopecommon = sprintf("%.2f",$slopecommon);
#				$sumXsqtotal = sprintf("%.2f",$sumXsqtotal); $sumXYtotal = sprintf("%.2f",$sumXYtotal); $sumYsqtotal = sprintf("%.2f",$sumYsqtotal);
				$var2=sprintf("%.2f",$var2); $var1=sprintf("%.2f",$var1); $totalSStotal = sprintf("%.2f",$totalSStotal);
				$residualSStotal = sprintf("%.2f",$residualSStotal); $residualDFtotal = sprintf("%d",$residualDFtotal);
				$slopetotal = sprintf("%.2f",$slopetotal); $intercepttotal = sprintf("%.2f",$intercepttotal); $Rsquaretotal = sprintf("%.2f",$Rsquaretotal);
				if ($Fstatistics ne "x") { $Fstatistics = sprintf("%.2f",$Fstatistics); }
				print "Pooled regression\t-\t-\t-\t".$residualSSpooled."\t".$residualDFpooled."\n";
				print "Common regression\t".$sumXsqcommon."\t".$sumXYcommon."\t".$sumYsqcommon."\t".$residualSScommon."\t".$residualDFcommon."\t".$slopecommon."\n";
#				print "Total regression\t".$sumXsqtotal."\t".$sumXYtotal."\t".$sumYsqtotal."\t".$residualSStotal."\t".$residualDFtotal."\t".$slopetotal.";".$intercepttotal.";".$Rsquaretotal."\n";
				print "Total regression\t".$var2."\t".$var1."\t".$totalSStotal."\t".$residualSStotal."\t".$residualDFtotal."\t".$slopetotal.";".$intercepttotal.";".$Rsquaretotal."\n";
				print "F-statistics\t".$Fstatistics."\n";
			}

			if ($Fstatistics ne "x") {
				# Read the F-statistics table
				my @statsvalues=();
				open (MYFILE1, $statsTable);
				while (<MYFILE1>) {
					$line = $_;
					$line =~ s/\r//g;	# remove linefeed
					$line =~ s/\n//g;	# remove linefeed
					push(@statsvalues, $line);
				}

				# Find the closest numerator DF across the columns and the closest denominator DF along the rows
				$numeratorDF = $residualDFcommon-$residualDFpooled;
				$denominatorDF = $residualDFpooled;
				$minidxCol=-1; $minidxRow=-1;
				@tmp1 = split(/\t/, $statsvalues[0]);
				for ($i=0; $i<scalar(@tmp1); $i++) {
					if ($tmp1[$i] =~ /^-?\d+\.?\d*$/) {
						if ($minidxCol == -1) {
							$minidxCol=0;
							$mindist=abs($tmp1[$i]-$numeratorDF);
						}
						else {
							if ($mindist > abs($tmp1[$i]-$numeratorDF)) {
								$minidxCol=$i;
								$mindist=abs($tmp1[$i]-$numeratorDF);
							}
						}
					}
				}
				for ($i=0; $i<scalar(@statsvalues); $i++) {
					@tmp1 = split(/\t/, $statsvalues[$i]);
					if ($tmp1[0] =~ /^-?\d+\.?\d*$/) {
						if ($minidxRow == -1) {
							$minidxRow=0;
							$mindist=abs($tmp1[0]-$denominatorDF);
						}
						else {
							if ($mindist > abs($tmp1[0]-$denominatorDF)) {
								$minidxRow=$i;
								$mindist=abs($tmp1[0]-$denominatorDF);
							}
						}
					}
				}

				# Get the F critical value
				@tmp1=split(/\t/,$statsvalues[$minidxRow]); $Fcritical=$tmp1[$minidxCol];
				if ($messageflag==1) { print "F-critical value[DF1,DF2]\t".$Fcritical."[".$numeratorDF.",".$denominatorDF."]\n"; }

				# Test for statistical significance
				if ($Fstatistics >= $Fcritical) { $significantflag=1; } else { $significantflag=0; }
#				print "[".$minidxCol.",".$minidxRow."]->[".$numeratorDF.",".$denominatorDF."]\t".$statsvalues[$minidxRow]."\n\n";
			}
			else { if ($messageflag==1) { print "F-critical value[DF1,DF2]\tx\n"; } }
		}
		else {
			$significantflag=-1; $Fstatistics=$Fcritical=$numeratorDF=$denominatorDF="x";
		}
	}
	return ($significantflag, $Fstatistics, $Fcritical, $numeratorDF, $denominatorDF);
}

sub permute {
	my $last = pop @_;
	unless (@_) {
	return map [$_], @$last;
	}
	return map { my $left = $_; map [@$left, $_], @$last } permute(@_);
}

sub combinations {
	return [] unless @_;
	my $first = shift;
	my @rest = combinations(@_);
	return @rest, map { [$first, @$_] } @rest;
}

sub testofproportion2samples {
	my($arrayRefX, $arrayRefY, $type) = @_;
	my @arrayA = @$arrayRefX;
	my @arrayB = @$arrayRefY;

	my $i;
	my @tmp1;

	my @data1;
	my @data2;
	my $n1;
	my $n2;
	my $avg1;
	my $avg2;
	my $z="";

	if (scalar(@arrayA)>0 && scalar(@arrayB) > 0) {

		# Create the proportion data sets
		for ($i=0,@data1=(); $i<scalar(@arrayA); $i++) {
			@tmp1 = split(/\t/, $arrayA[$i]);
			push(@data1, $tmp1[0]-$tmp1[1]);
		}
		for ($i=0,@data2=(); $i<scalar(@arrayB); $i++) {
			@tmp1 = split(/\t/, $arrayB[$i]);
			push(@data2, $tmp1[0]-$tmp1[1]);
		}

		# Assign the size of the data sets
		$n1 = scalar(@data1);
		$n2 = scalar(@data2);

		# Compute the average of the data sets
		for ($i=0,$avg1=0; $i<scalar(@data1); $i++) { $avg1 = $avg1 + $data1[$i]; }
		$avg1 = $avg1/scalar(@data1);

		for ($i=0,$avg2=0; $i<scalar(@data2); $i++) { $avg2 = $avg2 + $data2[$i]; }
		$avg2 = $avg2/scalar(@data2);

		# Compute the z-statistics
		$z = ($avg1-$avg2)/(($avg1*(1-$avg2))/$n1) + (($avg2*(1-$avg1))/$n2);
	}
	return $z;
}

# See Jerrold Zar, Biostatistical analysis, Chapter 23.1, pp486
sub twobyNcontingencyTable {
	my($arrayRef) = @_;
	my @obsarray = @$arrayRef;

	my $i;
	my $j;
	my $var1;
	my @tmp1;
	my @tmp2;

	my $errorflag=0;
	my @exparray=();
	my @rowcount=();
	my @columncount=();
	my $row; my $column; my $totalcount;

	my $chivalue=""; my $df="";

	if (scalar(@obsarray)==2) {
		# Set up the row and column count variables
		$row=2; for ($i=0; $i<$row; $i++) { push(@rowcount, 0); }
		@tmp1=split(/\t/, $obsarray[0]); $column=scalar(@tmp1); for ($i=0; $i<$column; $i++) { push(@columncount, 0); }
		$df = ($row-1)*($column-1);

		for ($i=0; $i<$row; $i++) {
			@tmp1 = split(/\t/, $obsarray[$i]);
			for ($j=0; $j<scalar(@tmp1); $j++) {
				$rowcount[$i] = $rowcount[$i] + $tmp1[$j];
				$columncount[$j] = $columncount[$j] + $tmp1[$j];
			}
		}

#		for ($i=0; $i<scalar(@rowcount); $i++) { print $rowcount[$i]."\n"; } print "\n";
#		for ($i=0; $i<scalar(@columncount); $i++) { print $columncount[$i]."\n"; }

		# Compute the total count (either sum the rowcount or columncount)
		for ($i=0,$totalcount=0; $i<scalar(@rowcount); $i++) { $totalcount = $totalcount + $rowcount[$i] }

		# Compute the expected frequencies
		for ($i=0; $i<$row; $i++) {
			for ($j=0,@tmp1=(); $j<scalar(@columncount); $j++) {
				push(@tmp1, (($rowcount[$i]/$totalcount)*$columncount[$j]));
			}
			push(@exparray, join("\t", @tmp1));
		}

#		for ($i=0; $i<scalar(@obsarray); $i++) { print $obsarray[$i]."\n" }
#		for ($i=0; $i<scalar(@exparray); $i++) { print $exparray[$i]."\n" }

 		for ($i=0,$chivalue=0; $i<scalar(@obsarray); $i++) {
			@tmp1 = split(/\t/, $obsarray[$i]);
			@tmp2 = split(/\t/, $exparray[$i]);
			if (scalar(@tmp1)==scalar(@tmp2) && scalar(@tmp1)>0) {
				for ($j=0; $j<scalar(@tmp1); $j++) {
#					print $j."\t".$tmp1[$j]."\t".$tmp2[$j]."\n";
					$chivalue = $chivalue + (($tmp1[$j]-$tmp2[$j])**2)/($tmp2[$j]);
				}
			}
			else { $errorflag=1; }
		}
#		if ($errorflag==0) { print $chivalue."\t".$df."\n\n"; }
	}
	return ($errorflag, $chivalue, $df);
}

sub createContingencyTable {
	my ($traitsRef, $dataRef, $targetidx, $refidx) = @_;
	my @traits=@$traitsRef; my @datatable=@$dataRef;

	my $i;
	my $var1;
	my @tmp1;
	my @tmp2;
	my @newTraits=(); my @newTargetCount=(); my @newRefCount=();

	$var1 = utilities::transposeMatrix(\@datatable);
	my @tranposeTable = @$var1;

#	print $tranposeTable[$targetidx]."\n";
#	print $tranposeTable[$refidx]."\n";

	my @contingencyTable=();
	@tmp1 = split(/\t/, $tranposeTable[$refidx]);
	@tmp2 = split(/\t/, $tranposeTable[$targetidx]);
#	print scalar(@traits)."\t".scalar(@tmp1)."\t".scalar(@tmp2)."\n";

	if (scalar(@traits)==scalar(@tmp1) && scalar(@traits)==scalar(@tmp2) && scalar(@traits)>0) {
		for ($i=0; $i<scalar(@tmp1); $i++) {
			if ($tmp1[$i]>0) {
#				print $traits[$i]."\t".$tmp1[$i]."\t".$tmp2[$i]."\n";
				push(@newTraits, $traits[$i]);
				push(@newRefCount, $tmp1[$i]);
				push(@newTargetCount, $tmp2[$i]);
			}
		}
		push(@contingencyTable, join("\t", @newRefCount));
		push(@contingencyTable, join("\t", @newTargetCount));
	}
#	print scalar(@newTraits)."\t".scalar(@newRefCount)."\t".scalar(@newTargetCount)."\n";
#	for ($i=0; $i<scalar(@contingencyTable); $i++) { print $contingencyTable[$i]."\n"; }

	return (\@newTraits, \@contingencyTable);
}

sub getmedian {
	my ($dataRef) = @_;
	my @data=@$dataRef;

	my $idx;
	my $medianvalue;
	my @sortdata= sort {$a<=>$b} @data;

	if (scalar(@data)>1) {
		$idx=floor(scalar(@sortdata)/2);
		if ((scalar(@sortdata)%2)==0) { $medianvalue=0.5*($sortdata[$idx-1]+$sortdata[$idx]); }
		else { $medianvalue=$sortdata[$idx]; }
	}
	elsif (scalar(@data)==1) { $medianvalue=$data[0]; }
	else { $medianvalue=""}

	return $medianvalue;
}

sub getSummaryStatistics {
	my ($dataRef) = @_;
	my @data=@$dataRef;
	my $avgvalue; my $stddev;

	my $i;
	my $sumX; my $sumXsq;

	if (scalar(@data)>0) {
		# Compute average
		for ($i=0,$sumX=0; $i<scalar(@data); $i++) { $sumX=$sumX+$data[$i]; }
		$avgvalue=$sumX/scalar(@data);
		# Compute standard deviation
		for ($i=0,$sumXsq=0; $i<scalar(@data); $i++) { $sumXsq=$sumXsq+($data[$i]-$avgvalue)*($data[$i]-$avgvalue); }
		$stddev=sqrt($sumXsq/scalar(@data));

	}
	return ($avgvalue,$stddev);
}

1;
